
require_relative 'level'
require_relative 'animated_object'
require_relative 'player'
require_relative 'wall'
require_relative 'pellet'
require_relative 'cage'
require_relative 'ghost'

module Pacman
  # DSL builder
  class LevelBuilder
    class << self
      def build(dsl)
        builder = new(dsl)
        builder.level
      end

      private :new
    end

    # builds a level according to the DSL
    def initialize(dsl)
      @level = Level.new
      @x = 0
      @y = 0

      # exec DSL
      instance_eval(dsl)

      # place ghosts to the cage
      if @level.cage_obj
        cage_obj = @level.cage_obj
        @level.ghosts = { blinky: BlinkyGhost.new(cage_obj.x, cage_obj.y, :left),
                          pinky: PinkyGhost.new(cage_obj.x + 1, cage_obj.y,
                                                :left),
                          inky: InkyGhost.new(cage_obj.x + 2, cage_obj.y, :right),
                          clyde: ClydeGhost.new(cage_obj.x + 3, cage_obj.y,
                                                :right) }
      end

      @level.pellets_left = @level.pellet_count

      @level
    end

    def resize(width, height)
      @level.resize(width, height)
    end

    def check_bounds(x, y)
      fail OutOfBoundsError.new if x < 0 || y < 0 || @level.width < x ||
          @level.height < y
    end

    # skip n columns if possible, otherwise skip row
    def skip(n)
      x = (@x + n) % @level.width
      y = (@y * @level.width + @x + n) / @level.width
      fail OutOfBoundsError.new if y >= @level.height || x >= @level.width

      @y = y
      @x = x
    end

    # go to specified position
    def goto(x, y)
      fail OutOfBoundsError.new if y < 0 || x < 0 || y >= @level.height ||
          x >= @level.width
      @x, @y = x, y
    end

    def place_object(clazz)
      check_bounds(@x, @y)
      @level[@y][@x] = clazz.new
      begin
        skip(1)
      rescue
        raise OutOfBoundsError.new unless @x == @level.width - 1 &&
              @y == @level.height - 1
      end
    end

    def wall
      place_object(Wall)
    end

    def pellet
      place_object(Pellet)
      @level.pellet_count += 1
    end

    def power_pellet
      place_object(PowerPellet)
    end

    # set player position
    def player(direction)
      check_bounds(@x, @y)
      player = Player.new(direction, @x, @y)
      level.player = player
      begin
        skip(1)
      rescue
        raise OutOfBoundsError.new unless @x == @level.width - 1 &&
          @y == @level.height - 1
      end
    end

    def cage
      check_bounds(@x, @y)
      cage_obj = Cage.new(@x, @y)
      @level[@y][@x] = cage_obj
      begin
        skip(1)
      rescue
        raise OutOfBoundsError.new unless @x == @level.width - 1 &&
            @y == @level.height - 1
      end

      @level.cage_obj = cage_obj
    end

    attr_reader :level
  end

  class OutOfBoundsError < StandardError
  end
end
